/* SPDX-License-Identifier: GPL-2.0-only */
#ifndef LABWC_IMG_XPM_H
#define LABWC_IMG_XPM_H

struct lab_data_buffer;

struct lab_data_buffer *img_xpm_load(const char *filename);

#endif /* LABWC_IMG_XPM_H */
