/* SPDX-License-Identifier: GPL-2.0-only */
#ifndef LABWC_SHOW_DESKTOP_H
#define LABWC_SHOW_DESKTOP_H

void show_desktop_toggle(void);
void show_desktop_reset(void);

#endif /* LABWC_SHOW_DESKTOP_H */
